# -*- coding: utf-8 -*-

import os.path
import re
from collections import OrderedDict
from ctypes import *

try:
	string_types = (unicode, str)
except NameError: # for python 3
	string_types = (str,)

from synthDriverHandler import SynthDriver, VoiceInfo, LanguageInfo, getSynth
from driverHandler import DriverSetting, StringParameterInfo, BooleanDriverSetting, NumericDriverSetting
import config
import addonHandler
import nvwave
import speech
import winKernel
from logHandler import log

addonHandler.initTranslation()

try:
	from synthDriverHandler import synthIndexReached, synthDoneSpeaking
except ImportError: # for NVDA below 2019.3
	import extensionPoints
	synthIndexReached = synthDoneSpeaking = extensionPoints.Action()

try:
	file_path = os.path.dirname(__file__.decode("mbcs"))
except AttributeError: # for python 3
	file_path = os.path.dirname(__file__)

abbreviationsLength = 4
isSpeaking = False
player = None

@WINFUNCTYPE(c_int, POINTER(c_char), POINTER(c_char), c_int)
def processAudio(udata, buffer, length):
	if not isSpeaking: return 1
	player.feed(string_at(buffer, length))
	return 0

@WINFUNCTYPE(None, c_int)
def onIndex(index):
	synth = getSynth()
	if index == -1:
		synthDoneSpeaking.notify(synth=synth)
	else:
		synthIndexReached.notify(synth=synth, index=index)
		synth.lastIndex = index

@WINFUNCTYPE(None)
def doneAudio():
	if isSpeaking:
		player.idle()

def freeLibrary(handle):
	if winKernel.kernel32.FreeLibrary(handle) == 0:
		raise WindowsError()
	return True

re_words = re.compile(r"\b(\w+)\b",re.U)
re_englishLetters = re.compile(r"\b([a-zA-Z])\b")
re_abbreviations = re.compile(u"\\b([\\d,bcdfghjklmnpqrstvwxzбвгджзклмнпрстфхцчшщ]{2,})\\b",re.U)
re_capAbbreviations = re.compile(u"([bcdfghjklmnpqrstvwxzбвгджзклмнпрстфхцчшщ]{3,})",re.U|re.I)
re_afterNumber = re.compile(r"(\d+)([^\.\:\-\/\!\?\d])")
re_omittedCharacters = re.compile(r"[\(\)\*_\"]+")
re_zeros = re.compile(r"\b\a?\.?(0+)")

ukrainianRules = {
	re.compile(u"\\b(й)\\s",re.U|re.I): U"й",
	re.compile(u"\\b(з)\\s",re.U|re.I): U"з",
	re.compile(u"\\s(ж)\\b",re.U|re.I): U"ж",
	re.compile(u"\\s(б)\\b",re.U|re.I): U"б",
	re.compile(u"'([яюєї])",re.I|re.U): u"ьй\\1",
	re.compile(u"ц([ьіяюєї])",re.U|re.I): U"тс\\1",
}

englishLetters = {
	'a': u"эй",
	'b': u"би",
	'c': u"си",
	'd': u"ди",
	'e': u"и",
	'f': u"эф",
	'g': u"джи",
	'h': u"эйчь",
	'i': u"ай",
	'j': u"джей",
	'k': u"кей",
	'l': u"эл",
	'm': u"эм",
	'n': u"эн",
	'o': u"оу",
	'p': u"пи",
	'q': u"къю",
	'r': u"ар",
	's': u"эс",
	't': u"ти",
	'u': u"ю",
	'v': u"ви",
	'w': u"да+блъю",
	'x': u"экс",
	'y': u"вай",
	'z': u"зэт",
}

russianLetters = {
	u"б": u"бэ",
	u"в": u"вэ",
	u"к": u"ка",
	u"с": u"эс",
	u"ь": u"мягкий знак",
	u"ъ": u"твёрдый знак",
}

englishPronunciation = {
	'x': u"кс",
	'ee': u"е е",
	'e': u"е",
	'y': u"ы",
	'j': u"дж",
}

pseudoEnglishPronunciation = {
	'x': u"кс",
	'ee': u"э э",
	'e': u"э",
	'y': u"ы",
	'j': u"дж",
}

#ukrainian to russian character map
#ukrainian soft "g" is not supported, becouse synth does not contain this phonem :(
ukrainianPronunciation = {
	u"и": u"ы",
	u"і": u"и",
	u"ї": u"ййи",
	u"е": u"э",
	u"є": u"е",
	u"ґ": u"г",
}

ukrainianPronunciationOrder = [u"и",u"і", u"ї", u"е", u"є", u"ґ"]

ukrainianLetters = {
	u"й": u"йот",
	u"ґ": u"Твэрдэ+ гэ",
	u"и": u"ы",
	u"і": u"и",
	u"ї": u"ййи",
	u"е": u"э",
	u"є": u"е",
}

letters = {}
letters.update(englishLetters)
letters.update(russianLetters)
russianZeros=[u"ноль ",u"ноля",u"нолей"]
ukrainianZeros=[u"нуль ",u"нулі ",u"нулів "]

def subZeros(match,zeros):
	l = len(match.group(1))
	if l == 1: return zeros[0]
	text = " " + str(l) + " "
	l = l%10
	if l == 1: return text+ zeros[0]
	elif l <5: return text+zeros[1]
	else: return text+zeros[2]

def expandAbbreviation(match):
	loweredText = match.group(1).lower()
	l = len(match.group(1))
	if (match.group(1).isupper() and (l <= abbreviationsLength and l > 1) and re_capAbbreviations.match(match.group(1))) or re_abbreviations.match(loweredText):
		expandedText = ""
		for letter in loweredText:
			expandedText += letters[letter] if letter in letters else letter
			if letter.isalpha(): expandedText+=" "
		return expandedText
	return loweredText

def subEnglishLetters(match):
	letter = match.group(1).lower()
	return englishLetters[letter]

def preprocessEnglishText(text, englishPronunciation):
	text = re_englishLetters.sub(subEnglishLetters, text)
	for s in englishPronunciation:
		text = text.replace(s, englishPronunciation[s])
	return text

def preprocessUkrainianText(text):
	for rule in ukrainianRules:
		text = rule.sub(ukrainianRules[rule],text)
	for s in ukrainianPronunciationOrder:
		text = text.replace(s, ukrainianPronunciation[s])
		#stupid python! replace() does not have ignore case, reg exprs also sucks
		text = text.replace(s.upper(), ukrainianPronunciation[s])
	return text

def processText(text,language, pseudoEnglishPronunciationFlag):
	if len(text) == 1:
		letter = text.lower()
		if language == "uk" and letter in ukrainianLetters: return ukrainianLetters[letter]
		elif letter in letters: return letters[letter]
		else: return letter
	text = re_omittedCharacters.sub(" ", text)
	text = re_zeros.sub(lambda match: subZeros(match,russianZeros if language=="ru" else ukrainianZeros),text)
	if language == "uk":
		text = preprocessUkrainianText(text)
	text = re_words.sub(expandAbbreviation,text) #this also lowers the text
	text = preprocessEnglishText(text, pseudoEnglishPronunciation if pseudoEnglishPronunciationFlag == True else englishPronunciation)
	text = re_afterNumber.sub(r"\1-\2", text)
	return text

class SynthDriver(SynthDriver):
	name = "newfon"
	description = "Newfon"

	supportedSettings = (
		SynthDriver.VoiceSetting(),
		SynthDriver.LanguageSetting(),
		SynthDriver.RateSetting(),
		DriverSetting("accel", _("&Acceleration"), True),
		DriverSetting("samplesPerSec", _("&Samples per second (hz)"), True),
		SynthDriver.PitchSetting(),
		SynthDriver.InflectionSetting(10),
		SynthDriver.VolumeSetting(),
		NumericDriverSetting("pauseBetweenPhrases", _("&Pause between phrases"), True),
		BooleanDriverSetting("useSynthDict", _("Use built-in &dictionary synthesizer (only for Russian language)"), True, defaultVal=True),
		BooleanDriverSetting("pseudoEnglishPronunciation", _("Include pseudo &english pronunciation"), True, defaultVal=True),
	)

	supportedCommands = {speech.IndexCommand, speech.PitchCommand}
	supportedNotifications = {synthIndexReached, synthDoneSpeaking}

	@classmethod
	def check(cls):
		return True

	def calculateMinMaxPitch(self,pitch,inflection):
		min,max=self.pitchTable[int(self.voice)]
		i=max-min
		i=int((i/50.0)*((inflection-50)/2))
		min-=i
		max+=i
		i=int((pitch-50)/1.3)
		min+=i
		max+=i
		return min,max

	def __init__(self):
		global player
		self._volume = 100
		self._language = "ru"
		self._pitch = 50
		self._accel = "0"
		self._pauseBetweenPhrases = 50
		self._useSynthDict = True
		self._pseudoEnglishPronunciation = True
		self._inflection = 50
		self.pitchTable=[(90,130),(190,330),(60,120),(220,340)]
		self.lastIndex = None

		player = nvwave.WavePlayer(channels=1, samplesPerSec=int(self.samplesPerSec), bitsPerSample=8, outputDevice=config.conf["speech"]["outputDevice"])
		self.dictdb_lib = WinDLL(os.path.join(file_path, "dictdb.dll"))
		self.dict_lib = WinDLL(os.path.join(file_path, "ndict.dll"))
		self.newfon_lib = WinDLL(os.path.join(file_path, "newfon_nvda.dll"))
		if not self.newfon_lib.initialize():
			raise RuntimeError("Newfon: initialization error")
		self.newfon_lib.set_callbacks(processAudio, onIndex, doneAudio)
		self.newfon_lib.set_dictionary(True)

	def terminate(self):
		global player
		self.cancel()
		player.close()
		player = None
		self.newfon_lib.terminate()
		try:
			freeLibrary(self.newfon_lib._handle)
			freeLibrary(self.dict_lib._handle)
			freeLibrary(self.dictdb_lib._handle)
		except WindowsError:
			log.exception("Can not unload dll")
		finally:
			del self.newfon_lib
			del self.dict_lib
			del self.dictdb_lib

	def speak(self, speechSequence):
		global isSpeaking
		isSpeaking = True
		textList = []
		for item in speechSequence:
			if isinstance(item, string_types):
				textList.append(processText(item, self._language,self._pseudoEnglishPronunciation))
			elif isinstance(item, speech.IndexCommand):
				self._speak(textList, item.index)
				textList = []
			elif isinstance(item, speech.PitchCommand):
				if textList:
					self._speak(textList)
					textList = []
				min, max = self.calculateMinMaxPitch(self._pitch+item.offset, self._inflection)
				self.newfon_lib.set_pitch_min(min)
				self.newfon_lib.set_pitch_max(max)
			elif isinstance(item, speech.SpeechCommand):
				log.debugWarning("Unsupported speech command: %s" % item)
			else:
				log.error("Unknown speech: %s" % item)
		self._speak(textList)

	def _speak(self, textList, index=-1):
		text = "".join(textList)
		text = text.encode("cp1251", errors="replace")
		self.newfon_lib.speakText(text, index)
		self.newfon_lib.set_pause(self._pauseBetweenPhrases)

	def cancel(self):
	
		global isSpeaking
		self.newfon_lib.cancel()
		isSpeaking = False
		player.stop()

	def _set_voice(self, value):
		self.newfon_lib.set_voice(int(value))
		self._set_pitch(self._pitch)

	def _get_voice(self):
		return str(self.newfon_lib.get_voice())

	def _get_availableVoices(self):
		voices = [_("male 1"), _("female 1"), _("male 2"), _("female 2")]
		return OrderedDict((str(index), VoiceInfo(str(index), name)) for index, name in enumerate(voices))

	def _set_volume(self, value):
		self._volume = value
		self.newfon_lib.set_volume(value)

	def _get_volume(self):
		return self._volume

	def _set_rate(self, value):
		self.newfon_lib.set_rate(value)

	def _get_rate(self):
		return self.newfon_lib.get_rate()

	def _set_pitch(self, value):
		self._pitch = value
		min,max=self.calculateMinMaxPitch(self._pitch,self._inflection)
		self.newfon_lib.set_pitch_min(min)
		self.newfon_lib.set_pitch_max(max)

	def _get_pitch(self):
		return self._pitch

	def pause(self, switch):
		player.pause(switch)

	def _set_language(self, language):
		self._language = language
		self.useSynthDict = self._useSynthDict

	def _get_language(self):
		return self._language

	def _get_availableLanguages(self):
		return OrderedDict((
			("ru", LanguageInfo("ru")),
			("uk", LanguageInfo("uk")),
		))

	def _set_inflection(self,inflection):
		self._inflection=inflection
		self._set_pitch(self._pitch)

	def _get_inflection(self):
		return self._inflection

	def _set_accel(self, value):
		self._accel = value
		self.newfon_lib.set_accel(int(value))

	def _get_accel(self):
		return self._accel

	def _get_availableAccels(self):
		return OrderedDict((str(x), StringParameterInfo(str(x), str(x))) for x in range(8))

	def _set_samplesPerSec(self, value):
		self.cancel()
		player.samplesPerSec = int(value)

	def _get_samplesPerSec(self):
		if player is not None:
			return str(player.samplesPerSec)
		return "10000"

	def _get_availableSamplespersecs(self):
		sampleRates = OrderedDict()
		for s in ("8000", "9025", "10000", "11025", "12000", "13025", "14000", "15025", "16000"):
			sampleRates[s] = StringParameterInfo(s, s)
		return sampleRates

	def _set_pauseBetweenPhrases(self, value):
		self._pauseBetweenPhrases = value
		self.newfon_lib.set_pause(self._pauseBetweenPhrases)

	def _get_pauseBetweenPhrases(self):
		return self._pauseBetweenPhrases

	def _set_useSynthDict(self, value):
		self._useSynthDict = value
		if self.language == "ru":
			self.newfon_lib.set_dictionary(value)
		else:
			self.newfon_lib.set_dictionary(False)

	def _get_useSynthDict(self):
		return self._useSynthDict

	def _set_pseudoEnglishPronunciation(self, value):
		self._pseudoEnglishPronunciation = value

	def _get_pseudoEnglishPronunciation(self):		
		return self._pseudoEnglishPronunciation